extern crate rand;

use std::cmp::Ordering;
use std::collections::HashMap;

use rand::{thread_rng, sample};

/// Alias for an L-System predecessor.
pub type Predecessor = (Option<String>, char, Option<String>);


/// A library that makes working with L-systems simple. A consequence
/// of this is that it only works with `char`s as predecessors and
/// `String`s as successors.
///
/// For a more general L-systems library,
/// see https://crates.io/crates/lindenmayer-system/.
pub struct LSystem<F: ?Sized> where F: Fn(&Vec<String>) -> String {

    /// The axiom for the L-system
    axiom: String,

    /// The current state of the L-system
    state: String,

    /// The L-system's set of productions. Each production is stored
    /// as a tuple with optional predecessor/successors for the production.
    rules: HashMap<Predecessor, Vec<String>>,

    /// Probability distributions for nondeterministic L-systems
    distributions: HashMap<Predecessor, Box<F>>,
}


impl<F: ?Sized> LSystem<F> where F: Fn(&Vec<String>) -> String {

    /// Creates a new L-system from the given axiom
    /// # Examples
    ///
    /// ```text
    /// use simple_lsystems::LSystem;
    /// let mut system: LSystem<Fn(&Vec<String>) -> String> = LSystem::with_axiom("b");
    /// ```
    pub fn with_axiom(axiom: &str) -> LSystem<F> {
        LSystem {
            axiom: axiom.into(),
            state: "".into(),
            rules: HashMap::new(),
            distributions: HashMap::new(),
        }
    }

    /// Adds a context-free rule. Exactly equivalent to calling
    /// `LSystem::add_context_rule` with the path and tree set to `None`.
    pub fn add_rule(&mut self, predecessor: char, successor: &str) {
        self.rules.entry((None, predecessor, None)).or_insert(Vec::new()).push(successor.into());
    }

    /// Adds a context-sensitive rule. `predecessor` is a 3-tuple
    /// where the first and third element are the path and tree, respectively,
    /// and the middle element is the predecessor.
    pub fn add_context_rule<T>(&mut self, predecessor: (Option<T>, char, Option<T>), successor: &str) where T: Into<String> {

        // Convert path from using &str to String
        let path = predecessor.0.map(|p| p.into());
        let tree = predecessor.2.map(|t| t.into());

        self.rules.entry((path, predecessor.1, tree)).or_insert(Vec::new()).push(successor.into());
    }

    /// Adds probability distribution for context-free predecessor.
    pub fn add_distribution(&mut self, predecessor: char, distribution: Box<F>) {
        self.distributions.insert((None, predecessor, None), distribution);
    }

    /// Adds probability distribution for context-sensitive predecessor.
    pub fn add_context_distribution(&mut self, predecessor: Predecessor, distribution: Box<F>) {
        self.distributions.insert(predecessor, distribution);
    }

    /// Resets the current state of the L-system so that the
    /// next state generated is equal to the axiom.
    pub fn reset(&mut self) {
        self.state = "".into();
    }

    /// Default distribution if one is unassigned.
    fn default(vals: &Vec<String>) -> String {
        // TODO: This could probably be more efficient.
        let mut rng = thread_rng();
        sample(&mut rng, vals.into_iter(), 1)[0].clone()
    }

    /// Since the HashMap is unordered, we need to manually sort
    /// the items when iterating over them to force the context-
    /// sensitive rules to be matched before the context-free rules.
    fn get_sorted_rules(&self) -> Vec<(&Predecessor, &Vec<String>)> {
        let mut sorted_rules: Vec<(&Predecessor, &Vec<String>)> = self.rules.iter().collect();

        sorted_rules.sort_by(|a, b| {
            let key_1 = a.0;
            let key_2 = b.0;

            match (key_1, key_2) {
                // key_1 is a context-sensitive rule and key_2 is context-free
                (&(Some(_), _, Some(_)), &(None, _, None)) => Ordering::Less,
                (&(Some(_), _, None   ), &(None, _, None)) => Ordering::Less,
                (&(None,    _, Some(_)), &(None, _, None)) => Ordering::Less,

                // key_1 is a context-free rule and key_2 is context-sensitive
                (&(None, _, None), &(Some(_), _, Some(_))) => Ordering::Greater,
                (&(None, _, None), &(Some(_), _, None   )) => Ordering::Greater,
                (&(None, _, None), &(None,    _, Some(_))) => Ordering::Greater,

                // Both keys are either context-free or context-sensitive, so the
                // order isn't particularly important.
                _ => key_1.cmp(key_2),
            }
        });

        sorted_rules
    }

    /// Gets the successor for a particular character. Accepts an index
    /// argument to do context-sensitive matching with. Since the rules
    /// are stored as keys in a hashmap with the predecessors/successors,
    /// this function will iterate over each key in the hashmap until it
    /// finds a match, instead of doing a normal `O(1)` lookup.
    fn get_successor(&self, ch: char, i: usize) -> Option<String> {
        for (key, value) in self.get_sorted_rules() {
            if key.1 != ch {
                continue
            }

            let (before, after) = self.state.split_at(i);

            // Look backwards in the string for the path
            let before_matches = match key.0 {
                Some(ref k) => before.to_string().chars().into_iter().rev().take(k.len()).collect::<String>() == *k,
                None => true,
            };

            // Look forwards in the string for the tree
            let after_matches = match key.2 {
                Some(ref k) => after.to_string().chars().into_iter().take(k.len()).collect::<String>() == *k,
                None => true,
            };

            // If we've got a match, use the distribution to select
            // a particular successor.
            if before_matches && after_matches {
                return Some(match self.distributions.get(&key) {
                    Some(f) => f(&value),
                    None => LSystem::<Fn(&Vec<String>) -> String>::default(&value),
                });
            }
        }

        None
    }
}


impl<F: ?Sized> Iterator for LSystem<F> where F: Fn(&Vec<String>) -> String {
    type Item = String;

    fn next(&mut self) -> Option<String> {
        let mut next = String::new();

        // The first iteration is just the axiom.
        if self.state == "" {
            self.state = self.axiom.clone();
            return Some(self.state.clone());
        }

        // Loop through each character in the current state
        // and find a replacement, or replace it with itself.
        for (i, ch) in self.state.chars().enumerate() {
            match self.get_successor(ch, i) {
                Some(successor) => {
                    next.push_str(&successor);
                    continue;
                },
                None => next.push(ch),
            };
        }

        self.state = next.clone();

        Some(next)
    }
}

/// Test that a basic example works
#[test]
fn basic_example() {
    let mut system: LSystem<Fn(&Vec<String>) -> String> = LSystem::with_axiom("b");
    system.add_rule('a', "ab");
    system.add_rule('b', "a");

    assert_eq!(system.next().unwrap(), "b");
    assert_eq!(system.next().unwrap(), "a");
    assert_eq!(system.next().unwrap(), "ab");
    assert_eq!(system.next().unwrap(), "aba");
    assert_eq!(system.next().unwrap(), "abaab");
    assert_eq!(system.next().unwrap(), "abaababa");
}

// Tests that a basic example of probability distributions works.
#[test]
fn test_distribution() {
    let mut system: LSystem<Fn(&Vec<String>) -> String> = LSystem::with_axiom("FG");
    system.add_rule('F', "F+");
    system.add_rule('F', "F-");
    system.add_rule('G', "G+");
    system.add_rule('G', "G-");

    system.add_distribution('F', Box::new(|vals: &Vec<String>| vals[0].clone()));
    system.add_distribution('G', Box::new(|vals: &Vec<String>| vals[1].clone()));

    assert_eq!(system.next().unwrap(), "FG");
    assert_eq!(system.next().unwrap(), "F+G-");
}

// Tests that closures and regular functions can be mixed and
// matched in the distributions hashmap.
#[test]
fn test_closures() {
    let mut system: LSystem<Fn(&Vec<String>) -> String> = LSystem::with_axiom("F");
    system.add_rule('F', "F+");

    fn distribution(vals: &Vec<String>) -> String {
        vals[0].clone()
    }

    system.add_distribution('F', Box::new(|vals: &Vec<String>| vals[0].clone()));
    system.add_distribution('G', Box::new(distribution));

    assert_eq!(system.next().unwrap(), "F");
    assert_eq!(system.next().unwrap(), "F+");
}

// Tests that `LSystem.reset` properly re-initializes the `LSystem`.
#[test]
fn test_reset() {
    let mut system: LSystem<&Fn(&Vec<String>) -> String> = LSystem::with_axiom("F");
    system.add_rule('F', "FF");

    assert_eq!(system.next().unwrap(), "F");
    assert_eq!(system.next().unwrap(), "FF");
    assert_eq!(system.next().unwrap(), "FFFF");

    system.reset();

    assert_eq!(system.next().unwrap(), "F");
    assert_eq!(system.next().unwrap(), "FF");
    assert_eq!(system.next().unwrap(), "FFFF");
}

// Tests that a basic context-sensitive example works.
#[test]
fn test_context() {
    let mut system: LSystem<&Fn(&Vec<String>) -> String> = LSystem::with_axiom("baaaaaaa");
    system.add_rule('b', "a");
    system.add_context_rule((Some("b"), 'a', None), "b");

    assert_eq!(system.next().unwrap(), "baaaaaaa");
    assert_eq!(system.next().unwrap(), "abaaaaaa");
    assert_eq!(system.next().unwrap(), "aabaaaaa");
    assert_eq!(system.next().unwrap(), "aaabaaaa");
    assert_eq!(system.next().unwrap(), "aaaabaaa");
    assert_eq!(system.next().unwrap(), "aaaaabaa");
    assert_eq!(system.next().unwrap(), "aaaaaaba");
    assert_eq!(system.next().unwrap(), "aaaaaaab");
    assert_eq!(system.next().unwrap(), "aaaaaaaa");
}

// Tests that a context-sensitive rule takes precedence over
// a context-free rule.
#[test]
fn test_context_precedence() {
    let mut system: LSystem<&Fn(&Vec<String>) -> String> = LSystem::with_axiom("abcb");
    system.add_rule('b', "x");
    system.add_context_rule((Some("c"), 'b', None), "y");

    assert_eq!(system.next().unwrap(), "abcb");
    assert_eq!(system.next().unwrap(), "axcy");
}

/// Branching and contexts don't work yet.
#[test]
#[should_panic]
fn test_branching_context() {
    let mut system: LSystem<&Fn(&Vec<String>) -> String> = LSystem::with_axiom("ABC[DE][SG[HI[JK]L]MNO]");

    system.add_context_rule((Some("BC"), 'S', Some("G[H]M")), "X");

    assert_eq!(system.next().unwrap(), "ABC[DE][SG[HI[JK]L]MNO]");
    assert_eq!(system.next().unwrap(), "ABC[DE][XG[HI[JK]L]MNO]");
}
